//
// * MinGW build options
//
//  gcc -shared -o trayicon.dll trayicon.c m_pd.h "PATH_TO_PD/pd.dll" Gdi32.Lib -DMSW
//
// Gdi32.Lib is included in "Windows SDK" or "Platform SDK" available from Microsoft
//

#define _WIN32_WINNT	0x0501
#define _WIN32_IE		0x0501

// Number of icons to load from a single file
#define MAX_ICONS_PER_FILE 100

#include <windows.h>
#include <tchar.h>
#include "m_pd.h"

#define TRAYICON_EXTERN __declspec(dllexport) extern

static t_class *trayicon_class;
WNDCLASSEX wcx;


enum CUSTOM_WM {
	WM_TRAY_ACTION = WM_USER
};


// -------- Structs ------------------------------

// Generic Link list
typedef struct _trayiconll
{
	void *item;
	struct _trayiconll *next;
} trayiconll;


typedef struct _trayicon
{
	t_object       x_obj;
	char		   pad[4];
	HWND           hWnd;
	HANDLE         hThread;
	DWORD          threadId;
	NOTIFYICONDATA icon_data;
	trayiconll     *icons;
	trayiconll     *popups;
	DWORD          input_buf;
	HINSTANCE      hInstance;
	int            icon_cnt;
	t_outlet       *fout;
	CRITICAL_SECTION cs;
} t_trayicon;


// When sending messages from tray icon, WndProc wants
// to no pointer to the matching instance.
typedef struct _popupitem
{
	t_trayicon *x;
	t_float    val;
} popupitem;


#ifdef __cplusplus
extern "C" {
#endif


// ----------------- Generic linklist function ------------------------


int tray_icon_linklist_push(trayiconll **linklist, trayiconll *element)
{
	trayiconll **tmp = linklist;

	int cnt = 0;

	if (*tmp == NULL) {
		*tmp = element;
	} else {
		while (*tmp != NULL) {
			tmp = &((*tmp)->next);
			cnt++;
		}
		*tmp = element;
	}

	return cnt;
}


static void trayicon_linklist_clear(trayiconll *linklist)
{
	trayiconll *tmp;
	while (linklist != NULL) {

		tmp = linklist;
		linklist = linklist->next;

		free(tmp);
	}
}


// ------------------ Basic funcs -------------------------------------

static void update_icon(t_trayicon *x)
{
	Shell_NotifyIcon(NIM_MODIFY, &(x->icon_data));
}


static void create_popup(HWND hWnd, LPARAM lParam)
{
	POINT lp;
	GetCursorPos(&lp);

	HMENU hPopup = GetSubMenu(GetMenu(hWnd), 0);

	SetForegroundWindow(hWnd);
	TrackPopupMenu(hPopup, TPM_RIGHTALIGN | TPM_BOTTOMALIGN | TPM_LEFTBUTTON, lp.x, lp.y, 0, hWnd, NULL);
	PostMessage(hWnd, WM_NULL, 0, 0);
}


static void add_icon(t_trayicon *x, t_symbol *s)
{
	int i;
	HICON smIcons[MAX_ICONS_PER_FILE];
	UINT cnt;
	trayiconll *new_item;

	DWORD file_check = GetFileAttributesA(s->s_name);
	if ((file_check == -1) || (file_check & FILE_ATTRIBUTE_DIRECTORY)) {
		error("Icon file not found");
		return;
	}


	cnt = ExtractIconExA(s->s_name, 0, NULL, smIcons, MAX_ICONS_PER_FILE);

	if (cnt < 1) {
		error("Failed to load icon file");
		return;
	}


	x->icon_cnt += cnt;


	for (i=0; i<cnt; i++) {
		int cnt;
		new_item = (trayiconll *)malloc(sizeof(trayiconll));
		new_item->item = smIcons[i];
		new_item->next = NULL;

		cnt = tray_icon_linklist_push(&(x->icons), new_item);
		post("Icon loaded to bank [%d]", cnt);
	}
}



static void clear_icons(t_trayicon *x)
{
	x->icon_data.hIcon = NULL;
	update_icon(x);

	trayiconll *tmp = x->icons;
	while (tmp != NULL) {
		DestroyIcon(tmp->item);
		tmp = tmp->next;
	}

	trayicon_linklist_clear(x->icons);
	x->icons = NULL;

	x->icon_cnt = 0;
}


static void setup_icon(t_trayicon *x)
{
	HICON smIcons[1];
	UINT cnt;

	ZeroMemory(&(x->icon_data), sizeof(NOTIFYICONDATA));




	x->icon_data.cbSize = sizeof(NOTIFYICONDATA);
	x->icon_data.hWnd = x->hWnd;
	x->icon_data.uID = 0;
	x->icon_data.uTimeout = 10000;

	
	x->icon_data.hIcon = NULL;

	_tcscpy(x->icon_data.szTip, TEXT(""));

	x->icon_data.dwInfoFlags = NIIF_INFO;
	x->icon_data.uFlags = NIF_ICON | NIF_TIP | NIF_INFO | NIF_MESSAGE;
	x->icon_data.uCallbackMessage = WM_USER;

	Shell_NotifyIcon(NIM_ADD, &(x->icon_data));
}




// ----------------- Window handling ------------------------

LRESULT CALLBACK wndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

		case WM_TRAY_ACTION:
			switch (lParam) {
				case WM_MOUSEMOVE:
					break;

				case WM_LBUTTONDOWN:
				case WM_RBUTTONDOWN:
					create_popup(hWnd, lParam);
					break;
			}
			break;


		case WM_COMMAND:
			EnterCriticalSection(&(((popupitem *)wParam)->x->cs) );
			((popupitem *)wParam)->x->input_buf = ((popupitem *)wParam)->val;
			LeaveCriticalSection(&(((popupitem *)wParam)->x->cs) );
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
			break;

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
	return 0;
}


static void register_window_class()
{
	wcx.cbSize = sizeof(WNDCLASSEX);
	wcx.style = 0;
	wcx.cbClsExtra = 0;
	wcx.cbWndExtra = 0;
	wcx.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcx.hIconSm = LoadIcon(NULL, IDI_WINLOGO);
	wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcx.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcx.lpszMenuName  = NULL;
	wcx.lpszClassName = TEXT("TRAYICONWINDOW");
	wcx.hInstance = GetModuleHandle(NULL);
	wcx.lpfnWndProc = wndProc;

	RegisterClassEx(&wcx);
}


// Creates thread and message loop per instance
DWORD WINAPI trayicon_worker(t_trayicon *x)
{
	MSG msg;

	x->hInstance = GetModuleHandle(NULL);

	x->hWnd = CreateWindow(
					TEXT("TRAYICONWINDOW"),
					TEXT("dummy"),
					0, //WS_OVERLAPPEDWINDOW | WS_VISIBLE, // Make window visible for debug
					CW_USEDEFAULT, CW_USEDEFAULT,
					100, 100,
					HWND_DESKTOP, NULL, x->hInstance, NULL);

	setup_icon(x);

	MENUITEMINFO mii;
	HMENU hMenu = CreateMenu();
	HMENU hPopup = CreateMenu();

	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = MIIM_TYPE | MIIM_SUBMENU;
	mii.fType = MFT_STRING;
	mii.hSubMenu = hPopup;
	mii.dwTypeData = TEXT("dummy");

	InsertMenuItem(hMenu, 0, TRUE, &mii);

	SetMenu(x->hWnd, hMenu);

	while (GetMessage(&msg, NULL, 0, 0) != 0) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	Shell_NotifyIcon(NIM_DELETE, &(x->icon_data));

}


static void add_popup (t_trayicon *x, t_symbol *s, t_float f)
{
	if (f < 0.0f) {
		error("Value must be 0 or greater");
		return;
	}

	HMENU hPopup = GetSubMenu(GetMenu(x->hWnd), 0);

	popupitem *pi = (popupitem *)malloc(sizeof(popupitem));
	pi->x   = x;
	pi->val = f;

	MENUITEMINFO mii;
	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = MIIM_TYPE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.dwTypeData = s->s_name;
	mii.wID = (UINT)pi;
	InsertMenuItemA(hPopup, GetMenuItemCount(hPopup), TRUE, &mii);
}


static void clear_popups (t_trayicon *x)
{
	HMENU hPopup = GetSubMenu(GetMenu(x->hWnd), 0);
	int cnt = GetMenuItemCount(hPopup);
	while (cnt--) {
		RemoveMenu(hPopup, 0, MF_BYPOSITION);
	}
	trayicon_linklist_clear(x->popups);
	x->popups = NULL;
}



// ------------------- Object callbacks ---------------------

// Currently connects up to five symbols.
static void show_balloon (t_trayicon *x, t_symbol *s1, t_symbol *s2, t_symbol *s3, t_symbol *s4, t_symbol *s5)
{
	char buf[256];
	char *pos = buf;
	char *words[] = {s1->s_name, s2->s_name, s3->s_name, s4->s_name, s5->s_name, "\0"};

	strcpy(pos, words[0]);

	int word_no = 1;
	
	while (strlen(words[word_no]) > 0) {
		pos += strlen(words[word_no - 1]);
		*pos = ' ';
		pos++;
		strcpy(pos, words[word_no]);
		word_no++;
	}

	strcpy(x->icon_data.szInfo, buf);
	Shell_NotifyIcon(NIM_MODIFY, &(x->icon_data));
	strcpy(x->icon_data.szInfo, "");
}


static void set_tooltip(t_trayicon *x, t_symbol *s)
{
	strcpy(x->icon_data.szTip, s->s_name);
	update_icon(x);
}


void trayicon_clear(t_trayicon *x, t_symbol *s)
{
	if (strcmp(s->s_name, "icons") == 0) {
		clear_icons(x);
		return;
	}

	if (strcmp(s->s_name, "popups") == 0) {
		clear_popups(x);
		return;
	}
}


static void trayicon_float(t_trayicon *x, t_floatarg f)
{
	int i;
	trayiconll *tmp;
	int index = (int)f;

	if (x->icon_cnt < 1) {
		error("No icons registered yet");
		return;
	}

	if ((index < 0) || (index >= x->icon_cnt)) {
		error("Invalid icon index");
		return;
	}

	tmp = x->icons;
	while (index > 0){
		tmp = tmp->next;
		index--;
	}

	x->icon_data.hIcon = tmp->item;
	update_icon(x);
}


static void trayicon_bang(t_trayicon *x)
{
	EnterCriticalSection(&(x->cs));
	if (x->input_buf != -1) {
		outlet_float(x->x_obj.ob_outlet, x->input_buf);
		x->input_buf = -1;
	}
	LeaveCriticalSection(&(x->cs));
}





// -------------------------------------------------------------

// Destructor
static void trayicon_free(t_trayicon *x)
{
	PostMessage(x->hWnd, WM_DESTROY, (WPARAM)x, 0);
	clear_icons(x);
	DeleteCriticalSection(&(x->cs));
}




// Constructor
static void * trayicon_new(void)
{
	t_trayicon *x = (t_trayicon *)pd_new(trayicon_class);

	x->icons         = NULL;
	x->popups        = NULL;
	x->input_buf     = -1;
	x->icon_cnt      = 0;
	x->hThread       = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)trayicon_worker, (void *)x, 0, &(x->threadId));
	outlet_new(&x->x_obj, &s_float);

	InitializeCriticalSection(&(x->cs));
	return (void *)x;
}


TRAYICON_EXTERN void trayicon_setup(void)
{
	register_window_class();

	trayicon_class = class_new(gensym("trayicon"),
					(t_newmethod)trayicon_new,
					(t_method)trayicon_free,
					sizeof(t_trayicon),
					CLASS_DEFAULT,
					(t_atomtype)0);

	class_addmethod(
			trayicon_class,
			(t_method)add_icon,
			gensym("icon"),
			A_DEFSYMBOL,
			0);

	class_addmethod(
			trayicon_class,
			(t_method)trayicon_clear,
			gensym("clear"),
			A_DEFSYMBOL,
			0);

	class_addmethod(
			trayicon_class,
			(t_method)set_tooltip,
			gensym("tooltip"),
			A_DEFSYMBOL,
			0);

	class_addmethod(
			trayicon_class,
			(t_method)show_balloon,
			gensym("balloon"),
			A_DEFSYMBOL,
			A_DEFSYMBOL,
			A_DEFSYMBOL,
			A_DEFSYMBOL,
			A_DEFSYMBOL,
			0);

	class_addmethod(
			trayicon_class,
			(t_method)add_popup,
			gensym("popup"),
			A_DEFSYMBOL,
			A_DEFFLOAT,
			0);
	class_addfloat(trayicon_class, trayicon_float);

	class_addbang(trayicon_class, trayicon_bang);

}


#ifdef __cplusplus
}
#endif
